/*             CSE 687 Project 3 --- Sparse Matrix Class 

            Student Name: Gang Cheng SSN:223-61-3010               */

/*******************************************************************
 * matrix.hh ---- header file for the module of the member & frind *
 *		  functions for class Matrix                       *
 *******************************************************************/

#ifndef Matrix_HH
#define Matrix_HH

#include <iostream.h>     // cin, cout, cerr
#include <stdlib.h>      // exit
#include <string.h>      // strlen, strcat, strcpy, strncpy 

// link list class 
struct llist {
	int id;          // connection node id
	double value;    // connection weight
	llist *next;     // link to next connection list
	llist * search(int id_index); // search for link-cell
      	void deallocate_llist(void);  // link-cells destructor 
};

// graph node class
struct node {
	int id;          // node id
	llist *link;     // link to connections
	node *next;      // link to next node
	void deallocate_node(void);   // nodes destructor	
};

// Matrix class 
class Matrix {
   private:  
      int size;      // matrix size
      llist *lptr;   // link list -- connection for nodes
      node *nhead, *nptr;  // top pointor & working pointor

   public:
      	// Matrix m[10]  - define arrays of objects	
	Matrix(void) { size = 0; lptr = NULL; nhead = nptr = NULL; } 
	Matrix(int msize);  // constructor
	Matrix(Matrix& m);  // Matrix m2 = m1 - copy constructor
	void deallocate_matrix();   // explicit matrix destructor
	~Matrix(void) { this->deallocate_matrix(); } // destructor
	// unary operator - m3 = m2 = m1	 
	Matrix& operator= (Matrix& m); 
	// binary operator + --- addition of two matrixes
	friend Matrix operator+ (Matrix& m1, Matrix& m2);
	// binary operator * --- multiplication of two matrixes 
	friend Matrix operator* (Matrix& m1, Matrix& m2);
 	// overloading istream operator >>
	friend istream& operator>> (istream& istrm, Matrix& m); 
 	// overloading ostream operator << 
	friend ostream& operator<< (ostream& ostrm, Matrix& m);
	int matrix_size(void) { return size; } // matrix size 
};	

#endif Matrix_HH


/*             cse 687 Project 3 --- Sparse Matrix Class 

            Student Name: Gang Cheng SSN:223-61-3010               */

/*******************************************************************
 * Matrix.cc ---- the module of the member & frind functions       *
 *                for class Matrix                                 *
 *******************************************************************/

//#include "matrix.hh"
#include <stdio.h> // NULL
extern void error(char *);      // error handler

///////////// class Matrix member functions ///////////////

// ----- < Matrix :: Matrix(int msize) > --------------//

// - constructor: allocate space for nodes
Matrix :: Matrix(int msize)
{
#define MAXSIZE 5000
  if (msize < 0 || msize >MAXSIZE) 
    error("error: too big or small matrix size\n");
  
  node *ntmp = NULL;
  nhead = nptr = NULL;
  size = 0;
  while (size < msize) {
    size++;
    // allocate a new node
    if ((nptr = new node) == '\0')
      error("error: exhaust system memory\n"); 
    
    // initialize
    nptr->link = NULL;
    nptr->next = NULL; 
    nptr->id = size;

    if (ntmp == NULL) nhead = nptr; // pointer to the first node
    else ntmp->next = nptr;
    ntmp = nptr;
  };
#undef MAXSIZE
}

// ----< Matrix :: Matrix(Matrix& m) >----------------- //

// - copy constructs Matrix object from an existing object
Matrix :: Matrix(Matrix& m)
{
  node *ntmp = NULL;
  nhead = nptr = NULL;
  m.nptr = m.nhead;
  size = m.size;
  int n = m.size;
  while (n-- != 0) {
    // allocate a new node
    if ((nptr = new node) == '\0')
      error("error: exhaust system memory\n"); 
    
    // initialize
    nptr->link = NULL;
    nptr->next = NULL; 
    
    // copy node information
    nptr->id = (m.nptr)->id;
   
    // allocate and copy connctions for a node
    llist *ltmp = NULL;
    m.lptr = (m.nptr)->link;
    while (m.lptr != NULL) {
      // allocate a new link-cell
      if ((lptr = new llist) == '\0')
	error("error: exhaust system memory\n");
      lptr->next = NULL; // initialize 
      // copy link-cell information
      lptr->id = (m.lptr)->id;
      lptr->value = (m.lptr)->value;
      if (ltmp != NULL) // first cell is a special case
	ltmp->next = lptr;
      else 
	nptr->link = lptr;
      ltmp = lptr;
      m.lptr = (m.lptr)->next;
    };
    m.nptr = (m.nptr)->next;
    if (ntmp == NULL) nhead = nptr; // pointer to the first node
    else ntmp->next = nptr;
    ntmp = nptr;
  };
}

// ----< Matrix& Matrix :: operator= (Matrix& m) >--------- //

// - sets Matrix object equal to object m
Matrix& Matrix :: operator= (Matrix& m)
{ 
  if (m.size != size) error("error: incorrect matrix size\n"); 

  m.nptr = m.nhead;
  nptr = nhead;
  int n = size;
  while (n-- != 0) {

    if (m.nptr == NULL || nptr == NULL)  // bound checking
      error("error: fatal error\n");
    // copy node information
    nptr->id = (m.nptr)->id;
   

    // deallocate space of m's previous connection, if any
    (nptr->link)->deallocate_llist();

    // allocate and copy connctions for a node
    llist *ltmp = NULL;
    m.lptr = (m.nptr)->link;
    while (m.lptr != NULL) {
      // allocate a new link-cell
      if ((lptr = new llist) == '\0')
	error("error: exhaust system memory\n");
      lptr->next = NULL; // initialize 
      // copy link-cell information
      lptr->id = (m.lptr)->id;
      lptr->value = (m.lptr)->value;
      if (ltmp != NULL) // first cell is a special case
	ltmp->next = lptr;
      else 
	nptr->link = lptr;
      ltmp = lptr;
      m.lptr = (m.lptr)->next;
    };
    m.nptr = (m.nptr)->next;
    nptr = nptr->next;
  };
  return *this;
}

// ----<Matrix operator+ (Matrix& m1, Matrix& m2)>------------- //

// - addition of two Matrixs
Matrix operator+ (Matrix& m1, Matrix& m2)
{
  if (m1.size != m2.size) error("error: incorrect matrix size\n"); 

  Matrix temp(m1.size);  // use togather with unary operator = which
                         // copies this temporary immediately exit 
                         // from this function
  temp = m1;   // make temp identical to m1
  m2.nptr = m2.nhead;
  temp.nptr = temp.nhead;
  int n = m2.size;
  while (n-- != 0) {
    llist *ltmp;
    m2.lptr = (m2.nptr)->link;
    while (m2.lptr != NULL) {
      // search for a connection 
      if ((ltmp = ((temp.nptr)->link)->search((m2.lptr)->id)) != NULL)
	ltmp->value += (m2.lptr)->value;   // addition operation
      else {   // insert a new connection
	if ((ltmp = new llist) == '\0')
	  error("error: exhaust system memory\n");
	// copy connection information
	ltmp->id = (m2.lptr)->id;
	ltmp->value = (m2.lptr)->value;
	// link into new matrix
	ltmp->next = (temp.nptr)->link;
	(temp.nptr)->link = ltmp;
      };
      m2.lptr = (m2.lptr)->next;
    };
    temp.nptr = (temp.nptr)->next;
    m2.nptr = (m2.nptr)->next;
  };
  return temp;
}

// ----<Matrix operator* (Matrix& m1, Matrix& m2)>------------- //

// - multiplication of two Matrixs
Matrix operator* (Matrix& m1, Matrix& m2)
{
  
  if (m1.size != m2.size) error("error: incorrect matrix size\n"); 

  Matrix temp(m1.size);  // use togather with unary operator = which
                         // copies this temporary immediately exit 
                         // from this function
  m1.nptr = m1.nhead;
  temp.nptr = temp.nhead;
  (temp.nptr)->link = NULL;
  (temp.nptr)->id = (m1.nptr)->id;
  while (m1.nptr != NULL) {
    int m = 1;
    while (m <= m1.size) { 
      llist *ltmp;
      m1.lptr = (m1.nptr)->link;
      double sum = 0.0;
      while (m1.lptr != NULL) {
	int n2 = (m1.lptr)->id;
	m2.nptr = m2.nhead;   
	while (m2.nptr && (m2.nptr)->id != n2) 
	  m2.nptr = (m2.nptr)->next;
	if (m2.nptr == NULL) error("fatal error");
	// search for a connection 
	if ((ltmp = ((m2.nptr)->link)->search(m)) != NULL)
	  sum += ((m1.lptr)->value)*(ltmp->value);   // multiplication  
	                                             // operation
	m1.lptr = (m1.lptr)->next;
      };
      if (sum != 0.0) { // insert a new connection
	if ((ltmp = new llist) == '\0')
	  error("error: exhaust system memory\n");
	// copy connection information
	ltmp->id = m;
	ltmp->value = sum;
	// link into new matrix
	ltmp->next = (temp.nptr)->link;
	(temp.nptr)->link = ltmp;
	
      };
      m++;
    };
    temp.nptr = (temp.nptr)->next;
    m1.nptr = (m1.nptr)->next;
  };
  return temp;
}

// ----<istream& operator>> (istream& istrm, Matrix& m)>----- //

// - overloading istream operator >>
istream& operator>> (istream& istrm, Matrix& m)
{
  #define MAXSIZE 5000   // maxium size of a matrix
  m.deallocate_matrix();


  int n = 0;
  char ch = ' ';
  node *ntmp = NULL;
  m.size = 0;
  
  while (istrm) {
    if (! (istrm >> n)) continue; // skip blank lines on the bottom of file
    m.size++;
    if (n != m.size) error("error: incorrect input format(1)");
    // allocate a new node
    if ((m.nptr = new node) == '\0')
      error("error: exhaust system memory\n"); 
    
    (m.nptr)->id = n;
    (m.nptr)->link = NULL;
    istrm >>ch;
    if (ch == ':') {
      istrm >> ch;
      // input and create link_cells
      while (ch != ';') {
	istrm.putback(ch);
	// allocate a new link-cell
	if ((m.lptr = new llist) == '\0')
	  error("error: exhaust system memory\n");
	istrm >>(m.lptr)->id >> ch;
	if (ch != 'w') error("error: incorrect input format");
	istrm >>(m.lptr)->value >>ch;
	// insert into links
	(m.lptr)->next = (m.nptr)->link;
	(m.nptr)->link = m.lptr;
      };
    }
    else error("error: incorrect input format");
    // initialize node
    m.nptr->next = NULL; 
    // pointer to the first node
    if (ntmp == NULL) m.nhead = m.nptr; 
    else ntmp->next = m.nptr;
    ntmp = m.nptr;
    
  };

  if (!istrm.eof() || istrm.bad()) 
    error("error: incorrect input format");    

  return istrm;

  #undef MAXSIZE 
}

// ----<istream& operator<< (istream& istrm, Matrix& m)>---- //

// - overloading istream operator <<
ostream& operator<< (ostream& ostrm, Matrix& m)
{
  #define ZERO '0'          // print char. for symbol zero
  node *ntmp = NULL;
  int n = m.size;

  ostrm <<"   ";
  ostrm.width(10);

  for (n = 1; n <= m.size; ostrm << n++) ostrm.width(10);
  ostrm << "\n   ";
  for (n = m.size ; (n--)!=0; ostrm << "..") ostrm.width(10);
  ostrm << '\n';
  
  m.nptr = m.nhead;
  while (m.nptr != NULL) {
    ostrm.width(4);
    ostrm << (m.nptr)->id << ":";
    for (n = 1; n <= m.size; n++) {
      ostrm.width(10);
      if ((m.lptr = ((m.nptr)->link)->search(n)) != NULL)   
	ostrm << (m.lptr)->value;
      else
	ostrm << ZERO;
    };
    ostrm << '\n';
    m.nptr = (m.nptr)->next;
  };

  if (ostrm.bad()) 
    error("error: incorrect output");    

  //  ostrm << "\nNote: " << ZERO << " represent 0\n";
  return ostrm;
} 

// ----< void Matrix :: deallocate_matrix() >---- //

// explicit matrix destructor
void Matrix :: deallocate_matrix()
{
  if (size != 0) {
    nptr = nhead;
    while (nptr != NULL) { 
      (nptr->link)->deallocate_llist();  
      nptr = nptr->next;
    };       
    nhead->deallocate_node();
  };
  size = 0;
  nhead = nptr = NULL;

  return;
}

// ----< void llist :: deallocate_llist() >---- //

// explicit link list destructor
void llist :: deallocate_llist()
{  
  llist *ltmp1, *ltmp2 = this;
  while (ltmp2 != NULL) {
    ltmp1 = ltmp2->next;	
    delete ltmp2; 
    ltmp2 = ltmp1;
  };
  return;
}

// ----< void node :: deallocate_node() >---- //

// explicit graph node destructor
void node :: deallocate_node()
{
  node *ntmp1, *ntmp2 = this;
  while (ntmp2 != NULL) {
    ntmp1 = ntmp2->next;	
    delete ntmp2; 
    ntmp2 = ntmp1;
  };
  return;
} 

// ----< llist * llist :: search(int id_index) >----//

// search for a connection in a link list
llist * llist :: search(int id_index)
{
  llist *ltmp = this;
  while ((ltmp != NULL) && (ltmp->id != id_index)) ltmp = ltmp->next;
  return (ltmp);
}


/*             CSE 687 Project 3 --- Sparse Matrix Class 

            Student Name: Gang Cheng SSN:223-61-3010               */

/*******************************************************************
 * proj3.cc ---- main program to test the member & frind functions *
 *               for class Matrix                                  *
 *******************************************************************/

#include <fstream.h> 
#include "matrix.hh"  // class Matrix

//--- <void error(char *s)> ---//
/* error handler */
void error(char *s)
{
  cerr << s; 
  exit(1);
} 

//--- < main program > ---//
/* test the member and friend functions for class Matrix */
main(int argc, char *argv[])
{
  char fname1[10], fname2[10];

  switch (argc) {
  case 1:
    cerr << "Input file name for matrix 1: ";
    cin >> fname1;
    cerr << "Input file name for matrix 2: ";
    cin >> fname2;
    break;
  case 2:
    strcpy(fname1,argv[1]);
    cerr << "Input file name for matrix 2: ";
    cin >> fname2;
    break;
  case 3:
    strcpy(fname1,argv[1]);
    strcpy(fname2,argv[2]);
    break;
  default:
    error("error: incorrect command line input\n");
    break;
  };
  
  ifstream fp1(fname1);
  ifstream fp2(fname2);
 
  // check i/o status
  if (!fp1) {
    cerr << "cannot open file: " << fname1 << "\n";
    exit(1);
  };
  if (!fp2) {
    cerr << "cannot open file: " << fname2 << "\n";
    exit(1);
  };

  // status information     
  cerr << "\nThese are files " << fname1 << " and " << fname2 << "\nProcessed by " <<  argv[0] << "\n\n";

  Matrix m1,m2;
  fp1 >> m1; fp2 >> m2;
//  int size = m1.matrix_size();
  Matrix m3(m2.matrix_size());
  cout << "\nm1 = \n"<<m1 << "\nm2 = \n" <<m2;
  cout << "\nm1 + m2 = \n" << m1 + m2;
  m3 = m1 * m2;  
  cout << "\nm1 * m2 =  \n" << m3;
  cout << "\nm1 + m1*m2  =  \n" << m1 + m1*m2;

  exit(0);

}
// ****** END of the program ******** //


// remove the following lines out into 2 seperate file2 (e.g. m1.dat, m2.dat)

/*********  Attached 2 sample matrix input files with the format:   ***********

   node_key_id1 : connected_node_id1 w weight connected_node_id2 w weight ... ;
   node_key_id1 : connected_node_id1 w weight connected_node_id2 w weight ... ;
     .......
     .......
   node_key_idn : connected_node_id1 w weight connected_node_id2 w weight ... ;

  *****************************************************************************/

/***** m1.dat *****

1 : 2w1.56 4w0.23; 
2 : ;
3 : 1w 2.4 3w 2.3;
4 : 2w23.4  1w4.45;  
5 : 2w1.56  4w0.23; 
6 : ;
7 : 3w2.4 7w2.3;
8 : 6w23.4  2w4.9;
9 : 1w2.4  8w0.1;
10 : 10w23.4;
11 : 2w-1.2 3w0.01;

****** end of m1.dat *******/

/***** m2.dat *****

1 : 1w2.45;
2 : 3w4.567 2w1.23;
3 : 6w23.4  8w4.452345; 
4 : 2w23.4  1w4.45;  
5 : 2w1.56 4w0.23; 
6 : 1w1.56 8w0.23; 
7 : ;
8 : 5w2.45 1w2.567 7w2.489;
9 : 1w1.2;
10 : ;
11 : 11w2.3;

****** end of m1.dat *******/

/***********************************************************
To run the test program, you can either enter 0 or 1 or 2 
input file name(s) on the command line, e.g. "proj3" or
"proj3 m1.dat" or "proj3 m1.dat m2.dat", in the case the program
needs more input file after starting the program, it will ask 
you to enter the file name(s). The output of the program will 
be on the standard output device(screen) by default. You can 
redirct the output to a file, e.g. "proj3 >proj3.out" or
"proj3 m1.dat >proj3.out" or "proj3 m1.dat m2.dat >proj3.out".  
************************************************************/
 

   
